#define QT_THREAD_SUPPORT
#include <qthread.h>

#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

//#include "/usr/src/rtl-moxa/include/mcua.h"
#include "mcua.h"

#include "exchange.h"
#include "tools.h"
#include "messages.h"

extern QMutex	dataMutex;
extern bool	withDevice;

namespace exch {

// lobushka vybrosov:
	bool vybros = false;

	static const char* DEV_SIGNAL_ABBR[] = {"ACK","NAK","NOD","ACN","ACY","ACW",
					"DSC","BUS","COL"};  
	static unsigned char DEV_SIGNAL_CODE[] = {ACK,NAK,NOD,ACN,ACY,ACW,
					DSC,BUS,COL};
	static int N_DEV_SIGNAL = 9;


	static unsigned char respbuf[MSG_SIZE];
	static unsigned char *msgdata;

	static int id_out = 0;
	static int id_stat = 0;
	static int id_data = 0;
	static int id_proc = 0;

	FILE*		srv_file;
	static int	n_read_data;
	static int	block_number;

	int	_maxBlockSize;	// yesli prishel udlinennyi paket, on
				// ne dolzhen byt' dlinneye e
	int	_blockSize0;	// data + HEAD of block

	static int	n_counters = 1;
	Counter*	counters[MAX_N_COUNTERS];
	bool		timeout_flag = false;
	bool		data_cycle_condition = true;
}

void exch::wait( int msec ) {
	if( msec > S2MS ) {
		QString str_add;
		QString str = "Wait " +
			str_add.setNum( msec/S2MS ) + " sec...";
		showMessage( str, msec );
        }
	ulong mksec = ulong(msec*MS2MKS);
	usleep( mksec );
}

void exch::terminate_program( void ) {

	jdebug( "Terminate programm" );

	close( id_out );
	close( id_data );
	close( id_stat );
	close( id_proc );

	free( msgdata );
}

void exch::initExchange( void ) {

	jdebug( "start of initThreads:" );

	for( int i=0; i<MAX_N_COUNTERS; i++ ) counters[i]=NULL;
	srv_file = NULL;
	n_read_data = 0;
	block_number = 512;


	msgdata = (unsigned char *) malloc(SIZE_STREAM_FIFO*sizeof(uchar));

	id_proc = open( "/proc/moxa/mcua0/device", O_RDWR ); // nomera blockov
	if( (id_data = open( "/dev/rtf12", O_RDONLY | O_NONBLOCK )) < 0 ) { //  	// 
		perror( "?Open fifo" );
		exit(3);
	}
	id_stat = open( "/dev/rtf10", O_RDONLY ); 	// 

	ioctl( id_data, IO_FIFO_FLUSH,0 );


	if( (id_out = open( "/dev/mcua0",O_RDWR )) < 0 ) { // 
		perror( "?Open mcua" );
		exit(3);
	}
	
	atexit( terminate_program );

}

void exch::setCounters( Counter** p_counters ) {
	for( int i=0; i<MAX_N_COUNTERS; i++ ) {
		counters[i] = p_counters[i];
	}
}

void exch::setSrvFile( FILE* f ) {
	srv_file = f;
}

int exch::sendCommand( const char* command,
			bool save=true ) {
	char buf[30];
	int ibuf[4];
	int retval;

	int key = sscanf( command,"%X %X %X %X",&ibuf[0],
                                        &ibuf[1],&ibuf[2],&ibuf[3] );
	if( key > 0 ) {
		for( int i=0; i<key; i++ ) buf[i] = ibuf[i]; // Yavno privodim tipy
		if( save && srv_file ) {
			fprintf( srv_file,"-> %X",(uchar)buf[0] );
			for( int i=1; i<key; i++ ) fprintf( srv_file," %X",(uchar)buf[i] );
		}
		retval = write( id_out, buf, key ); // 
		if( retval < 0 ) { 
			perror("txd: ");
			switch( errno ) {
//				case -87	: Protocol...
				case ETIMEDOUT	: retval = DSC; break;
				case EBUSY	: retval = BUS; break;
				case ENOSPC	: debug( "No space left on device mcua0" );
							exit(1);
				case ECONNREFUSED : retval = COL; break;
				default	: debug( "!!! unknown errno=%d",errno ); 
							if( srv_file ) fprintf( srv_file, "Unknown errno=%d\n",errno );
							exit(1);
			}
		}
		if( retval > 0x80 ) {	// Signal
			respbuf[0] = (uchar)retval;
			for( int i=0; i<N_DEV_SIGNAL; i++ ) {
				if( respbuf[0] == DEV_SIGNAL_CODE[i] ) {
					if( save && srv_file ) fprintf( srv_file,"   <- %s\n",DEV_SIGNAL_ABBR[i] );
					if( respbuf[0] == DSC ) return(-2);
					return(0);
				}
			}
			if( srv_file ) fprintf( srv_file,"   <- Unknown signal: %X\n",respbuf[0] );
			return(0);
		} else {
			if( retval >= MSG_SIZE ) retval = MSG_SIZE - 1;
			respbuf[0] = retval+1;
			int yy = read( id_stat, respbuf+1, retval );
			if( yy <= 0 ) { 
				perror("read id_stat: ");
			}
			if( save && srv_file ) {
				fprintf( srv_file,"   <-... " );
				for( int i=1; i<respbuf[0]; i++ ) {
					fprintf( srv_file,"%X ",respbuf[i] );
				}
				fprintf( srv_file,"\n" );
			}
			return(1);
		}
	}
	return(-1);
} // end of namespace

int exch::sendCommand( char address,
			char command,
			unsigned char** r,
			int buf_len=2,
			char low=0,
			char high=0,
			bool save=true ) {
	char	buf[30];
	*r = respbuf;

	if( buf_len > 4 ) buf_len = 4;
	else if( buf_len < 2 ) buf_len = 2;
	buf[0] = address;
	buf[1] = command;
	buf[2] = low;
	buf[3] = high;
		
	if( save && srv_file ) {
		fprintf( srv_file,"-> %X %X",
		(uchar)address,(uchar)command );
		if( buf_len > 2 ) fprintf( srv_file," %X",(uchar)low );
		if( buf_len > 3 ) fprintf( srv_file," %X",(uchar)high );
	}
	int retval = write( id_out, buf, buf_len ); // 
	if( retval < 0 ) { 
		switch( errno ) {
			case ETIMEDOUT	: retval = DSC; break;
			case EBUSY	: retval = BUS; break;
			case ENOSPC	: debug( "No space left on device mcua0" );
						exit(1);
			case ECONNREFUSED : retval = COL; break;
			default	: debug( "!!! unknown errno=%d", errno );
					if( srv_file ) fprintf( srv_file, "Unknown errno=%d\n",errno );
					exit(1);
		}
	}
	if( retval > 0x80 ) {	// Signal
		respbuf[0] = (uchar)retval;
		for( int i=0; i<N_DEV_SIGNAL; i++ ) {
			if( respbuf[0] == DEV_SIGNAL_CODE[i] ) {
				if( save && srv_file ) fprintf( srv_file,"   <- %s\n",DEV_SIGNAL_ABBR[i] );
				if( respbuf[0] == DSC ) return(-2);
				return(0);
			}
		}
		if( srv_file ) fprintf( srv_file,"   <- Unknown signal: %X\n",respbuf[0] );
		return(0);
	} else {
		if( retval >= MSG_SIZE ) retval = MSG_SIZE - 1;
		respbuf[0] = retval+1;
		int yy = read( id_stat, respbuf+1, retval );
		if( yy <= 0 ) { 
			perror("read id_stat: ");
		}
		if( save && srv_file ) {
			fprintf( srv_file,"   <-... " );
			for( int i=1; i<respbuf[0]; i++ ) {
				fprintf( srv_file,"%X ",respbuf[i] );
			}
			fprintf( srv_file,"\n" );
		}
		return(1);
	}
}

QString exch::talk( const QString& text ) {
	QString response="";
	int resp = sendCommand( text.data() );
	if(  resp == 1 ) {
		for( int i=1; i<respbuf[0]; i++ ) {
			QString str;
			str.sprintf( "%X ",respbuf[i] );
			response.append( str );
		}
		return response;
	} else if( (resp == 0) || (resp == -2) ) {
		for( int i=0; i<N_DEV_SIGNAL; i++ ) {
			if( respbuf[0] == DEV_SIGNAL_CODE[i] ) {
				response.sprintf( "%s",DEV_SIGNAL_ABBR[i] );
				return( response.data() );
			}
		}
		response.sprintf( "Unknown signal: %X", respbuf[0] );
		return response;
	} else {
		response = "Invalid command!";
		return response;
	}
}

void exch::*read_data( void ) {
	jdebug( "+++++++++ read_data: Start..." );
	n_read_data = 0;

	int count_byte = 0;
	while( data_cycle_condition && withDevice ) {
		while( n_read_data < n_counters*block_number ) {
			if( !data_cycle_condition
					|| !withDevice ) break;
			count_byte = read( id_data, msgdata,1 );
			if( count_byte == 0 ) {
				usleep( 10000L );
				continue;
			}
			int size = (int)msgdata[0];
			if( size > _maxBlockSize ) size = _maxBlockSize;
			read( id_data, msgdata+1, size );

			putData();
			n_read_data++;
		}
		n_read_data = 0;
		dataMutex.unlock();
	}
	dataMutex.unlock();
	int res = 7;
	return( &res );
}

int exch::setBlockSize( int size ) {
	int blockSize = size;
	_blockSize0 = size+HEAD_SIZE;
	_maxBlockSize = _blockSize0 + HEAD_SIZE; //dlia udlinennykh paketov
	if( _maxBlockSize >= SIZE_STREAM_FIFO ) {
		_maxBlockSize = SIZE_STREAM_FIFO-1;
		_blockSize0 = _maxBlockSize - HEAD_SIZE;
		blockSize = _blockSize0-HEAD_SIZE;
	}
	return blockSize;
}

void exch::setBlockNumber( int number ) {
	block_number = number;
}

int exch::getNblock( void ) {
	if( !timeout_flag ) return n_counters*block_number;
	else return n_read_data;
}

void exch::setNcounters( int number ) {
	if( number <= MAX_N_COUNTERS ) n_counters = number;
}

int exch::getBlockNumber( void ) {
	return block_number;
}

int exch::getNreadData( void ) {
	return n_read_data;
}

void exch::setTimeout( bool tmout ) {
	timeout_flag = tmout;
}

bool exch::isTimeout( void ) {
	return timeout_flag;
}

void exch::startCounting( void ) {
	n_read_data = 0;
}

void exch::quitDataCycle( void ) {
	jdebug( "quitDataCycle" );
	data_cycle_condition = false;
	dataMutex.unlock();
}

void exch::timeoutAction( void ) {
	timeout_flag = true;
	dataMutex.unlock();
}

void exch::waitData( void ) {
	if( !dataMutex.locked() ) dataMutex.lock(); // if unlock
	dataMutex.lock();
}

void exch::writeToProc( char* to_proc_file, int size ) {
	write( id_proc,to_proc_file,size );
}

void exch::putData( void ) {
	Counter *currentCounter = NULL;
	for( int i=0; i<n_counters; i++ ) {
		if( counters[i] )
			if( msgdata[2] == (uchar)counters[i]->getModule() )
				currentCounter = counters[i];
	}

	if( !currentCounter ) {
		if( srv_file ) fprintf( srv_file, "Unknown counter with address %X\n",msgdata[2] );
		return;
	}

	uchar size = msgdata[0];

	bool ok = true;
	if( size != (_blockSize0-1) ) {
		ok = false;
	}

	int _blnumber = int(msgdata[3] + (msgdata[4] << 8));

	int dnumb = abs( _blnumber - currentCounter->blockCounter.get() );
	if( dnumb != 0 ) {
		if( srv_file ) fprintf( srv_file,"%s block %02X-1 %02X skipped\n",
				currentCounter->getName(),msgdata[3],msgdata[4] );
		currentCounter->blockCounter.set( _blnumber );
	}

// LOVUSHKA vybrosov:
	if( currentCounter->getModule()
			== counters[0]->getModule() ) {
		vybros = false;
	}
	if( ok ) {
		currentCounter->addBlockLabel( _blnumber );
		for( uchar t=HEAD_SIZE; t<=size; t=t+2 ) {
			count_t val = msgdata[t]+(msgdata[t+1] << 8);
			currentCounter->putData( val );
//	Yest' vybros???
			if( currentCounter->getModule()
				== counters[0]->getModule() ) {
				if( val >= 1000 ) vybros = true;
			}
		}
		if( vybros && srv_file ) {
			for( int t=0; t<=size; t++ ) {
				fprintf( srv_file,"%02X ",msgdata[t] );
			}
			fprintf( srv_file,"\n" );
		}
		currentCounter->blockCounter.incr();
	} else {
		jdebug( "!!!!!!!!!!!!!!%s: Skip block !!!!!!!",
				currentCounter->getName() );
		if( srv_file ) fprintf( srv_file,"%s Skip block %02X %02X\n",
				currentCounter->getName(),msgdata[3],msgdata[4] );
	}
}
