#include <qmessagebox.h>
#include "names.h"
#include "power.h"
#include "exchange.h"
#include "trberror.h"
#include "tools.h"
#include "messages.h"
#include "cfg.h"


Power::Power( void ) {
	_name = "HV power";
	_safety = -1;
	_currentVoltage = -1;
	ready = FALSE;
}

Power::~Power() {
}

void Power::loadConst( void ) {
	jdebug( "power: LoadConst" );

	_module		= leafDataToChar( deviceCfgFile,_name,"Common","Module" );
	_getStatus	= leafDataToChar( deviceCfgFile,_name,"Command","GetStatus" );
	_getIdent	= leafDataToChar( deviceCfgFile,_name,"Command","GetIdent" );
	_getCrc		= leafDataToChar( deviceCfgFile,_name,"Command","GetCrc" );
	_setVoltage	= leafDataToChar( deviceCfgFile,_name,"Command","SetVoltage" );
	_getVoltage	= leafDataToChar( deviceCfgFile,_name,"Command","GetVoltage" );
	_highOn		= leafDataToChar( deviceCfgFile,_name,"Command","HighON" );
	_highOff	= leafDataToChar( deviceCfgFile,_name,"Command","HighOFF" );
	_safetyOn	= leafDataToChar( deviceCfgFile,_name,"Command","SafetyON" );
	_safetyOff	= leafDataToChar( deviceCfgFile,_name,"Command","SafetyOFF" );
	_getTemper	= leafDataToChar( deviceCfgFile,_name,"Command","GetTemper" );
	_getConst	= leafDataToChar( deviceCfgFile,_name,"Command","GetConst" );
	_reset		= leafDataToChar( deviceCfgFile,_name,"Command","Reset" );

	 const char* _id = leafDataToStr( deviceCfgFile,_name,"Common","Identification" )
				->data();
	_nIdent = sscanf( _id,"%X%X%X%X", &_ident1,&_ident2,&_ident3,&_ident4 );

	ready = true;
	jdebug( "power: end of LoadConst" );
}

int Power::get( char _module, char _command, const char* _name,
		const char* _warning, unsigned char** p_respbuf, int _dialog ) {
	QString str;
	int i;
	int ret;
	int resp;

	for( i=0; i<N_REPEAT_COMMAND; i++ ) {
		resp = exch::sendCommand( _module,_command,p_respbuf );
		if( resp == 1 ) return(1);	// data
		exch::wait( DEVICE_WAIT_INTERVAL );
	}
	if( (*p_respbuf)[0] == DSC ) {
		str.sprintf( "%s disconnected!",_name );
		ret = -1;
	} else {
		str.sprintf( "%s: %s",_name,_warning );
		ret = -2;
	}
	if( !_dialog ) return(ret);
	switch ( showWarning( str ) ) {
		case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return ret;
		case  QMessageBox::Ignore: return(ret);
		case  QMessageBox::Retry : return(0);
		default : return(0);
	}
}

int Power::set( char _module, char _command, const char* _name,
		const char* _warning, int _dialog ) {
	QString str;
	int resp;
	int i;
	unsigned char *respbuf;

	while( 1 ) {
		for( i=0; i<N_REPEAT_COMMAND; i++ ) {
			resp = exch::sendCommand( _module,_command,&respbuf );
			if( respbuf[0] == ACY ) return(1);
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !_dialog ) return(0); 	
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return(0); 
			case  QMessageBox::Retry : continue; 
		}
	}
}

int Power::set_hvOn( char _module, char _command, const char* _name,
		const char* _warning, int _dialog ) {
	QString str;
	int resp;
	int i;
	unsigned char *respbuf;

	while( 1 ) {
		for( i=0; i<N_REPEAT_COMMAND; i++ ) {
			resp = exch::sendCommand( _module,_command,&respbuf );
			if( respbuf[0] == ACY ) return(1);
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( respbuf[0] == ACW ) return(-2);
		if( !_dialog ) return(0); 	
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return(0); 
			case  QMessageBox::Retry : continue; 
		}
	}
}

int Power::set( char _module, char _command, char _data, const char* _name,
			const char* _warning, int _dialog ) {
	QString str;
	int resp;
	unsigned char *respbuf;
	int i;

	while( 1 ) {
		for( i=0; i<N_REPEAT_COMMAND; i++ ) {
			resp = exch::sendCommand( _module,_command,&respbuf,3,_data );
			if( respbuf[0] == ACY ) return(1);
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !_dialog ) return(0); 	
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}	
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return(0); 
			case  QMessageBox::Retry : continue; 
		}
	}
}

int Power::set( char _module, char _command, char _data_low, char _data_high,
			const char* _name, const char* _warning, int _dialog ) {
	QString str;
	int resp;
	unsigned char *respbuf;
	int i;

	while( 1 ) {
		for( i=0; i<N_REPEAT_COMMAND; i++ ) {
			resp = exch::sendCommand( _module,_command,&respbuf,4,_data_low,_data_high );
			if( respbuf[0] == ACY ) return(1);
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !_dialog ) return(0); 	
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}	
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return(0); 
			case  QMessageBox::Retry : continue; 
		}
	}
}


int Power::getIdent( void ) {
	QString str;
	unsigned char *respbuf;

	while( 1 ) {
		str.sprintf( "Module \"%s\"",_name );
		switch ( get( _module, _getIdent, str.data(),
			"Error in getIdent!", &respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
			if( (respbuf[0] != _nIdent+1 ) ||
				(_ident1 != respbuf[1]) || (_ident2 != respbuf[2]) ||
				(_ident3 != respbuf[3]) || (_ident4 != respbuf[4]) ) {
					str.sprintf( "Module \"%s\": Invalid identification: %X %X %X %X!",
						_name,respbuf[1],respbuf[2],respbuf[3],respbuf[4] );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return(0);
						case  QMessageBox::Retry : break;
					}
				} else return(1);

			}
		}
	}
}

int Power::reset( void ) {
	QString str;
	int resp;
	unsigned char* respbuf;
	int i;
	int ret=0;

	jdebug( "Power::reset:" );
	_safety = -1;
	_currentVoltage = -1;
	int cycle = 1;
	while( cycle ) {
		for( i=0; i<N_REPEAT_COMMAND; i++ ) {
			resp = exch::sendCommand( _module, _reset, &respbuf );
			if( resp == -2 ) {	// disconnect
				return(1);
			}
		}
		str.sprintf( "Module \"%s\": Error in Reset!",_name );
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); cycle=0; ret=0; break;
			case  QMessageBox::Ignore: cycle=0; ret=0; break;
			case  QMessageBox::Retry : break;
		}
	}
	return( ret );
}

int Power::getConst( void ) {
	QString str;
	unsigned char *respbuf;

	while( 1 ) {
		str.sprintf( "Module \"%s\"",_name );
		switch ( get( _module, _getConst, str.data(),
				"Error in getConst!",&respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
				if( respbuf[0] == 5 ) {
					_const1 = (int)respbuf[1];
					_const2 = (int)respbuf[2];
					_const3 = (int)respbuf[3];
					_const4 = (int)respbuf[4];
					return(1);
				}
				str.sprintf( "Module \"%s\": Error in getConst!",_name );
				switch ( showWarning( str ) ) {
					case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
					case  QMessageBox::Ignore: return(0);
					case  QMessageBox::Retry : break;
				}
			}
		}
	}
}

int Power::testCrc( void ) {
	QString str;
	unsigned char *respbuf;

	jdebug( "testCrc:" );
	while( 1 ) {
		str.sprintf( "Module \"%s\"",_name );
		switch ( get( _module, _getCrc, str.data(),
				"Error in getCrc!", &respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
				if( (respbuf[0] != 2) || (respbuf[1] != 0) ){
					str.sprintf( "Module \"%s\": EEPROM corrupted!",_name );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return(0);
						case  QMessageBox::Retry : break;
					}
				} else return(1);
			}
		}
	}
}

int Power::getStatus( uchar& _status, int _dialog ) {
	QString str;
	unsigned char *respbuf;

//	jdebug( "getStatus:" );
	while( 1 ) {
		str.sprintf( "Module \"%s\"",_name );
		switch ( get( _module, _getStatus, str.data(),
				"Error in getStatus!",&respbuf,_dialog ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
				if( respbuf[0] != 2 ) {
					if( !_dialog ) return(0);
					str.sprintf( "Module \"%s\": Invalid status!", _name );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return(0);
						case  QMessageBox::Retry : continue;
					}
				} else  { _status = respbuf[1]; return(1); }
			}
		}
	}
}

int Power::setVoltage( int _hvLev, int _dialog, bool _wait=true ) {
	QString str;
	int _hvLev_;
	int _dataHV,_dataHV_;
	unsigned char* respbuf;

	jdebug( "setVoltage:" );
	if( _currentVoltage == _hvLev ) return(1);
//	_dataHV = 255 + _const2 -
//		(int)(_hvLev*(double)(255 + _const2 - _const1)); 

	_dataHV = (int)((double)_hvLev * 0.001 * (double)_const1) - _const2;
	jdebug( "_dataHV = %d",_dataHV );
	if( _dataHV < MIN_HV_VALUE ) _dataHV = MIN_HV_VALUE;
	if( _dataHV > MAX_HV_VALUE ) _dataHV = MAX_HV_VALUE;
	while( 1 ) {
		_currentVoltage = -1;
		str.sprintf( "Module \"%s\"",_name );
		if( set( _module, _setVoltage, (char)_dataHV,
			str.data(),"Error in setVoltage!", _dialog ) == 0 ) {
			return(0);
		}
		switch( get( _module, _getVoltage, str.data(),
			"Error in getVoltage!",&respbuf, _dialog ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
				if( respbuf[0] == 2 ) {
					if( _const1 == 0 ) {
						_hvLev_ = -1;
					} else {
						_dataHV_ = (int)respbuf[1];
						_hvLev_ = (int)(1000.0*(double)(_const2+_dataHV_)
							/ (double)_const1);
						if( _hvLev != 0 ) {
							if( fabs((double)(_hvLev-_hvLev_))/
								(double)_hvLev <= VALUE_TOLERANCE ) {
								_currentVoltage = _hvLev;
								if( _wait ) exch::wait( HV_WAIT_INTERVAL ); // msec
								return(1);
							}
						} else if( _hvLev_ == 0 ) {
							_currentVoltage = 0;
							return(1);
						}
					}
					str.sprintf( "Module \"%s\": Error in setVoltage: %d != %d!",
						_name, _hvLev, _hvLev_ );
				} else {
					str.sprintf( "Module \"%s\": Error in setVoltage: Numb. of ret. val. != 2!",
						_name );
				}
				if( !_dialog ) return(0);
				switch ( showWarning( str ) ) {
					case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
					case  QMessageBox::Ignore: return(0);
					case  QMessageBox::Retry : break;
				}
			}
		}
	}
}

int Power::getTemper( double& _temper, int _dialog ) {
	QString str;
	int _high, _low, _dataTemper;
	unsigned char* respbuf;

//	jdebug( "getTemper:" );
	while( 1 ) {
		str.sprintf( "Module \"%s\"",_name );
		switch( get( _module, _getTemper, str.data(),
			"Error in getTemper!",&respbuf, _dialog ) ) {
			case  0 : continue;
			case -1 : return(-1);
			case -2 : return(-2);
			case  1 : {
				if( respbuf[0] == 3 ) {
					_low = (int)respbuf[1];
					_high = (int)respbuf[2];
					_dataTemper = (_high << 8) + _low;
					_temper = -103.0 + (double)_dataTemper / 4.0;
					return(1);
				} else {
					str.sprintf( "Module \"%s\": Error in getTemper: Numb. of ret. val. != 3!",
						_name );
				}
				if( !_dialog ) return(0);
				switch ( showWarning( str ) ) {
					case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
					case  QMessageBox::Ignore: return(0);
					case  QMessageBox::Retry : break;
				}
			}
		}
	}
}

int Power::turnHV( const QString& _on, int _dialog ) {
	int _on_int = 0;
	if( operator==( _on.upper(),"ON" ) ) _on_int = ON;
	if( operator==( _on.upper(),"OFF" ) ) _on_int = OFF;
	return( turnHV( _on_int,_dialog ) );
}

int Power::turnHV( int _on, int _dialog ) {
	int ret;
	if( _on == ON ) ret = setHVOn( _dialog );
	else ret = setHVOff( _dialog );
	return( ret );
}

int Power::setHVOn( int _dialog ) {
	int ret;
	QString str;
	int _lock_hv;

	jdebug( "setHVOn:" );
//	if( isHVOn() ) return(1);
	if( _safety ) {
		setSafetyOff( _dialog );
		setSafetyOn( _dialog );
	}

RETRY:
	str.sprintf( "Module \"%s\"",_name );
	ret = set_hvOn( _module,_highOn,str.data(),
		"Error in setHVOn!",_dialog );
	if( ret == 0 ) return(0);
	exch::wait( HV_LOCK_WAIT ); // 100 millisec
	_lock_hv = isHVLock();
	if( (_lock_hv == 0) && (ret == 1) ) return(1); // It's OK
	if( !_dialog ) return(0);

	if( _lock_hv != 1 ) {
		str.sprintf( "Module \"%s\": Error in setHVOn!",_name );
		switch ( showWarning( str ) ) {
			case  QMessageBox::Retry : goto RETRY;
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return(0);
			default : return(0);
		}
	}

//	OVER_LIGHT && SAFETY_ON:

	str = "Overlight! High Voltage turned off!";
	switch ( showStrongWarning( str,"Ignore","Correct..." ) ) {
		case  QMessageBox::Ignore: return(0);
		case  QMessageBox::Abort : {
			str = "Remove light pollution!";
			switch( showStrongWarning( str,"Ignore","Set HV On" ) ) {
				case	QMessageBox::Ignore: return(0);
				case	QMessageBox::Abort : break;
			}
			setHVOff( _dialog );
			setSafetyOff( _dialog );
			setSafetyOn( _dialog );
			goto RETRY;
		}
		default : return(0);
	}
}

int Power::setHVOff( int _dialog ) {
	int ret;
	QString str;

	jdebug( "setHVOff:" );
//	if( isHVOn() ) {
	str.sprintf( "Module \"%s\"",_name );
	ret = set( _module,_highOff,str.data(),
		"Error in setHVOff!",_dialog );
//	}
	return( ret );
}

int Power::turnSafety( const QString& _on, int _dialog ) {
	int _on_int = 0;
	if( _on.upper() == "ON" ) _on_int = ON;
	if( _on.upper() == "OFF" ) _on_int = OFF;
	return( turnSafety( _on_int,_dialog ) );
}

int Power::turnSafety( int _on, int _dialog ) {
	int ret;
	if( _on ) ret = setSafetyOn( _dialog );
	else ret = setSafetyOff( _dialog );
	return( ret );
}

int Power::setSafetyOn( int _dialog ) {
	int ret=1;
	QString str;

	jdebug( "setSafetyOn:" );
	if( _safety != ON ) {
		str.sprintf( "Module \"%s\"",_name );
		ret = set( _module,_safetyOn,str.data(),
			"Error in setSafetyOn!",_dialog );
		if( ret == 1 ) _safety = ON;
	}
	return( ret );
}

int Power::setSafetyOff( int _dialog ) {
	int ret=1;
	QString str;

	jdebug( "setSafetyOff:" );
	if( _safety != OFF ) {
		str.sprintf( "Module \"%s\"",_name );
		ret = set( _module,_safetyOff,str.data(),
			"Error in setSafetyOff!",_dialog );
		if( ret == 1 ) _safety = OFF;
	}
	return( ret );
}

int Power::testOverLight( int _dialog ) {
	QString str;
	int ret;

//	jdebug( "testOverLight!" );
	while( 1 ) {
		ret = isHVLock();
		if( ret == 0 ) return(1);
		if( !_dialog ) return(0);
		if( ret == -1 ) {
			str.sprintf( "Module \"%s\" : Invalid status!",_name );
			switch ( showWarning( str ) ) {
				case  QMessageBox::Ignore: break;
				case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); break;
				case  QMessageBox::Retry : continue;
			}
		} else {
			jdebug( "power: OverLight!!!" );
			str = "Overlight! High Voltage turned off!";
			switch ( showStrongWarning( str,"Continue","Exit" ) ) {
				case  QMessageBox::Ignore: break;
				case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); break;
			}
			jdebug( "power: setHVOff()" );
			setHVOff( _dialog );
			jdebug( "power: setSafetyOff()" );
			setSafetyOff( _dialog );
			jdebug( "power: setSafetyOn()" );
			setSafetyOn( _dialog );
		}
		return(0);
	}
}

int Power::isOverLight( void ) {
	uchar _status;

	if( getStatus( _status,0 ) != 1 ) return(-1);
	if( _status & OVR_LIGHT ) return(1);
	else return(0);
}

int Power::isHVLock( void ) {
	uchar _status;

	if( getStatus( _status,0 ) != 1 ) return(-1);
	if( _status & HV_LOCK ) return(1);
	else return(0);
}

int Power::isHVOn( void ) {
	uchar _status;

	if( getStatus( _status,0 ) != 1 ) return(-1);
	if( _status & HV_ON ) {
		jdebug( "HV is ON" );
		return(ON);
	}
	else {
		jdebug( "HV is OFF" );
		return(OFF);
	}
}
