#include <qmessagebox.h>
#include "names.h"
#include "tracker.h"
#include "exchange.h"
#include "tools.h"
#include "messages.h"
#include "trberror.h"
#include "cfg.h"

Tracker::Tracker( void ) {
	_name = "Tracker";
	ready = false;
}

Tracker::~Tracker() {
}

void Tracker::loadConst( void ) {
	jdebug( "tracker: LoadConst" );

	_module		= leafDataToChar( deviceCfgFile,_name,"Common","Module" );
	_getStatus	= leafDataToChar( deviceCfgFile,_name,"Command","GetStatus" );
	_getCrc		= leafDataToChar( deviceCfgFile,_name,"Command","GetCrc" );

	_shiftRA	= leafDataToChar( deviceCfgFile,_name,"Command","ShiftRA" );
	_shiftDec	= leafDataToChar( deviceCfgFile,_name,"Command","ShiftDec" );
	_moveRA		= leafDataToChar( deviceCfgFile,_name,"Command","MoveRA" );
	_moveDec	= leafDataToChar( deviceCfgFile,_name,"Command","MoveDec" );
	_stopMotion	= leafDataToChar( deviceCfgFile,_name,"Command","StopMotion" );
	_testMotion	= leafDataToChar( deviceCfgFile,_name,"Command","TestMotion" );
	_getIdent	= leafDataToChar( deviceCfgFile,_name,"Command","GetIdent" );
	_getConst	= leafDataToChar( deviceCfgFile,_name,"Command","GetConst" );
	_reset		= leafDataToChar( deviceCfgFile,_name,"Command","Reset" );

	const char* _id = leafDataToStr( deviceCfgFile,_name,"Common","Identification" )
				->data();
	_nIdent = sscanf( _id,"%X%X%X%X", &_ident1,&_ident2,&_ident3,&_ident4 );

	ready = true;
}

int Tracker::get( char _module, char _command, const char* _name,
		const char* _warning, unsigned char** p_respbuf, int dialog ) {

	for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
		int resp = exch::sendCommand( _module,_command,p_respbuf );
		if( resp == 1 ) return 1;	// data
		exch::wait( DEVICE_WAIT_INTERVAL );
	}
	QString str = _name;
	int ret = 0;
	if( (*p_respbuf)[0] == DSC ) {
		str += " disconnected!";
		ret = -1;
	} else {
		str += ": "; str += _warning;
		ret = -2;
	}
	if( !dialog ) return ret;
	switch ( showWarning( str ) ) {
		case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return ret;
		case  QMessageBox::Ignore: return ret;
		case  QMessageBox::Retry : return 0;
		default : return 0;
	}
}

int Tracker::set( char _module, char _command, const char* _name,
		const char* _warning, int dialog ) {
	while( 1 ) {
		unsigned char *respbuf;
		for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
			exch::sendCommand( _module,_command,&respbuf );
			if( respbuf[0] == ACY ) return 1;
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !dialog ) return(0); 	
		QString str = _name;
		if( respbuf[0] == DSC ) {
			str += " disconnected!";
		} else {
			str += ": "; str += _warning;
		}
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return 0; 
			case  QMessageBox::Retry : continue; 
		}
	}
}

int Tracker::set( char _module, char _command, char _data, const char* _name,
			const char* _warning, int dialog ) {
	while( 1 ) {
		unsigned char *respbuf;
		for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
			exch::sendCommand( _module,_command,&respbuf,3,_data );
			if( respbuf[0] == ACY ) return 1;
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !dialog ) return 0; 	
		QString str;
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}	
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return 0; 
			case  QMessageBox::Retry : continue; 
		}
	}
}

int Tracker::set( char _module, char _command, char _data_low, char _data_high,
			const char* _name, const char* _warning, int dialog ) {
	while( 1 ) {
		unsigned char *respbuf;
		for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
			exch::sendCommand( _module,_command,&respbuf,4,_data_low,_data_high );
			if( respbuf[0] == ACY ) return 1;
			exch::wait( DEVICE_WAIT_INTERVAL );
		}
		if( !dialog ) return 0;
		QString str;
		if( respbuf[0] == DSC ) {
			str.sprintf( "%s disconnected!",_name );
		} else {
			str.sprintf( "%s: %s",_name,_warning );
		}	
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
			case  QMessageBox::Ignore: return 0; 
			case  QMessageBox::Retry : continue;
		}
	}
}

int Tracker::getIdent( void ) {
	jdebug( "Tracker::getIdent:" );
	while( 1 ) {
		QString str = "Module "+QString("\"")+_name+QString("\"");
		unsigned char *respbuf;
		switch( get( _module, _getIdent, str.data(),
			"Error in getIdent!", &respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return -1;
			case -2 : return -2;
			case  1 : {
			if( (respbuf[0] != _nIdent+1 ) ||
				(_ident1 != respbuf[1]) || (_ident2 != respbuf[2]) ||
				(_ident3 != respbuf[3]) || (_ident4 != respbuf[4]) ) {
					str.sprintf( "Module \"%s\": Invalid identification: %X %X %X %X!",
						_name,respbuf[1],respbuf[2],respbuf[3],respbuf[4] );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return 0;
						case  QMessageBox::Retry : break;
					}
				} else return 1;
			}
		}
	}
}

int Tracker::reset( void ) {
	jdebug( "Tracker::reset:" );
	int ret=0;
	int cycle=1;
	while( cycle ) {
		for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
			unsigned char* respbuf;
			int resp = exch::sendCommand( _module, _reset, &respbuf );
			if( resp == -2 ) {	// disconnect
				return 1;
			}
		}
		QString str;
		str.sprintf( "Module \"%s\": Error in Reset!",_name );
		switch ( showWarning( str ) ) {
			case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); cycle=0; ret=0; break;
			case  QMessageBox::Ignore: cycle=0; ret=0; break;
			case  QMessageBox::Retry : break;
		}
	}
	return ret;
}

int Tracker::getConst( void ) {
	while( 1 ) {
		unsigned char *respbuf;
		QString str = "Module "+QString("\"")+_name+QString("\"");
		switch( get( _module, _getConst, str.data(),
				"Error in getConst!",&respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return -1;
			case -2 : return -2;
			case  1 : {
				if( respbuf[0] == 5 ) {
					_const1 = (int)respbuf[1];
					_const2 = (int)respbuf[2];
					_const3 = (int)respbuf[3];
					_const4 = (int)respbuf[4];
					return 1;
				}
				str.sprintf( "Module \"%s\": Error in getConst!",_name );
				switch ( showWarning( str ) ) {
					case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0; 
					case  QMessageBox::Ignore: return 0;
					case  QMessageBox::Retry : break;
				}
			}
		}
	}
}

int Tracker::testCrc( void ) {
	jdebug( "testCrc:" );
	while( 1 ) {
		QString str = "Module "+QString("\"")+_name+QString("\"");
		unsigned char *respbuf;
		switch( get( _module, _getCrc, str.data(),
				"Error in getCrc!", &respbuf,1 ) ) {
			case  0 : continue;
			case -1 : return -1;
			case -2 : return -2;
			case  1 : {
				if( (respbuf[0] != 2) || (respbuf[1] != 0) ){
					str.sprintf( "Module \"%s\": EEPROM corrupted!",_name );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return 0;
						case  QMessageBox::Retry : break;
					}
				} else return 1;
			}
		}
	}
}

int Tracker::getStatus( uchar& status, int dialog ) {
	jdebug( "getStatus:" );
	while( 1 ) {
		QString str = "Module "+QString("\"")+_name+QString("\"");
		unsigned char *respbuf;
		switch( get( _module, _getStatus, str.data(),
				"Error in getStatus!",&respbuf,dialog ) ) {
			case  0 : continue;
			case -1 : return -1;
			case -2 : return -2;
			case  1 : {
				if( respbuf[0] != 2 ) {
					if( !dialog ) return 0;
					str.sprintf( "Module \"%s\": Invalid status!", _name );
					switch ( showWarning( str ) ) {
						case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0;
						case  QMessageBox::Ignore: return 0;
						case  QMessageBox::Retry : continue;
					}
				} else  { status = respbuf[1]; return 1; }
			}
		}
	}
}

char Tracker::toCharShift( double shift ) {
	double val = shift * _const1;
	jdebug( "toCharShift shift=%f val=%f",shift,val); 
	if( val > MAX_SIGN_CHAR ) val = MAX_SIGN_CHAR;
	else if( val < MIN_SIGN_CHAR ) val = MIN_SIGN_CHAR;
	return char(val);
}

int Tracker::shiftRA( double shift, int dialog ) {
	char shift_char = toCharShift(shift);
	jdebug( "shiftRA:%f char:%d",shift,shift_char );
RETRY:
	QString str = "Module "+QString("\"")+_name+QString("\"");
	if( set( _module,_shiftRA, shift_char,
			str.data(),"Error in shiftRA!", dialog ) == 0 ) {
		return 0;
	}
	for( int i=0; i<MAX_TRACKER_TEST_INTERVALS; i++ ) {
		switch( testMotion( dialog ) ) {
			case  0 : goto RETRY;
			case -1 : return -1;
			case -2 : return -2;
			case  2 : break;
			case  1 : return 1;
		}
		exch::wait( DEVICE_WAIT_INTERVAL );
	}
	stopMotion( dialog );
	if( !dialog ) return 0;
	str = "Infinite Motion!";
	switch( showStrongWarning( str ) ) {
		case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0; 
		case  QMessageBox::Ignore: return 0;
		default : return 0;
	}
	return 0;
}

int Tracker::shiftDec( double shift, int dialog ) {
	char shift_char = toCharShift(shift);
	jdebug( "shiftDec:%f char:%d",shift,shift_char );
RETRY:
	QString str = "Module "+QString("\"")+_name+QString("\"");
	if( set( _module,_shiftDec, shift_char,
			str.data(),"Error in shiftDec!", dialog ) == 0 ) {
		return 0;
	}
	for( int i=0; i<MAX_TRACKER_TEST_INTERVALS; i++ ) {
		switch( testMotion( dialog ) ) {
			case  0 : goto RETRY;
			case -1 : return -1;
			case -2 : return -2;
			case  2 : break;
			case  1 : return 1;
		}
		exch::wait( DEVICE_WAIT_INTERVAL );
	}
	stopMotion( dialog );
	if( !dialog ) return 0;
	str = "Infinite Motion!";
	switch( showStrongWarning( str ) ) {
		case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return 0; 
		case  QMessageBox::Ignore: return 0;
		default : return 0;
	}
	return 0;
}

int Tracker::moveRA( char direct, int dialog ) {
// "Beskonechnoe dvizhenie"
	jdebug( "moveRA%d" );
	QString str = "Module "+QString("\"")+_name+QString("\"");
	return set( _module,_moveRA, direct,
		str.data(),"Error in moveRA!", dialog );
}

int Tracker::moveDec( char direct, int dialog ) {
// "Beskonechnoe dvizhenie"
	jdebug( "moveDec%d" );
	QString str = "Module "+QString("\"")+_name+QString("\"");
	return set( _module,_moveDec, direct,
		str.data(),"Error in moveDec!", dialog );
}

int Tracker::testMotion( int dialog ) {
	unsigned char *respbuf;
	for( int i=0; i<N_REPEAT_COMMAND; i++ ) {
		exch::sendCommand( _module,_testMotion,&respbuf,2,0,0,false );
		if( respbuf[0] == ACY ) return 1;
		if( respbuf[0] == ACW ) return 2;
		exch::wait( DEVICE_WAIT_INTERVAL );
	}
	QString str = "Module "+QString("\"")+_name+QString("\"");
	int ret = 0;
	if( respbuf[0] == DSC ) {
		str += " disconnected!";
		ret = -1;
	} else {
		str += ": Error in testMotion!";
		ret = -2;
	}
	if( !dialog ) return ret; 	
	switch ( showWarning( str ) ) {
		case  QMessageBox::Abort : trberr::setCode(TRB_EXIT); return ret;
		case  QMessageBox::Ignore: return ret; 
		case  QMessageBox::Retry : return 0; 
	}
	return 0;
}

int Tracker::stopMotion( int dialog ) {
	jdebug( "stopMotion:" );
	QString str = "Module "+QString("\"")+_name+QString("\"");
	return( set( _module,_stopMotion,str.data(),
		"Error in stopMotion!",dialog ) );
}

/*
int Tracker::getStopCorrect( void ) {
	jdebug( "Tracker: currentApert = %d",_currentAperture ); 
	return( (int)_stopCorrect[ _currentAperture ] );
}

int Tracker::getStopCorrectNew( void ) {
	jdebug( "Tracker: currentApert = %d",_currentAperture ); 
	return( (int)_stopCorrect_new[ _currentAperture ] );
}

void Tracker::startCorrection( void ) {
	jdebug( "Tracker: Start Correction" );
	for( int i=0; i<_nStops; i++ ) {
		_stopCorrect_new[i] = _stopCorrect[i];
	}
}

void Tracker::acceptCorrection( int dialog ) {
	jdebug( "Tracker: Accept Correction" );
	for( int i=0; i<_nStops; i++ ) {
		_stopCorrect[i] = _stopCorrect_new[i];
	}
	loadStops( dialog );

// Copy to tree:
	QString _str,_num,_stopLeaf;
	for( int i=0; i<_nStops; i++ ) {
		_str = _stopName[i];
		_str.append( "\t" );
		_num.setNum( _stopFoclen[i],'f',2 );
		_str.append( _num );
		_str.append( " " );
		_num.setNum( _stopCorrect[i] );
		_str.append( _num );
		_stopLeaf = "Stop";
		_num.setNum( _stop[i] );
		_stopLeaf.append( _num );
		jdebug( " ________ _str = %s leaf=%s",_str.data(),
			_stopLeaf.data() );
		setLeafData( cfgFile,"Devices",_name,
			_stopLeaf.data(),_str );
	}
}
*/